# *****************************************************************
#
# Licensed Materials - Property of IBM
#
# (C) Copyright IBM Corp. 2017, 2020. All Rights Reserved.
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
#
# ******************************************************************

import numpy as np
import sys
import math
import warnings

from snapml.utils import _is_mpi_enabled
_mpi_enabled = _is_mpi_enabled()

if sys.version_info[0] < 3:
    import libsnapmllocal2 as libsnapml
else:
    import snapml.libsnapmllocal3 as libsnapml

class BoostingMachine:

    """
    BoostingMachine

    This class implements a boosting machine that can be used to construct an ensemble of decision trees.
    It can be used for both clasification and regression tasks.
    In constrast to other boosting frameworks, Snap ML's boosting machine dose not utilize a fixed maximal tree depth at each boosting iteration.
    Instead, the tree depth is sampled at each boosting iteration according to a discrete uniform distribution.
    The fit and predict functions accept numpy.ndarray data structures.

    Boosting params dictionary has the following structure:

    params = {
        'boosting_params' {
            'num_round': 10,
            'objective': 'mse',
            'random_state': 0,
            'colsample_bytree': 1.0,
            'subsample': 1.0,
            'verbose': False,
            'lambda_l2': 0.0,
            'early_stopping_rounds': 10,
            'compress_trees': False,
            'base_score': None
        },
        'tree_params': {
            'min_max_depth: 1,
            'max_max_depth: 6,
            'num_threads': 1,
            'use_histograms': True,
            'hist_nbins': 256,
            'use_gpu': False,
            'gpu_id': 0,
            'select_probability': 1.0
        },
        'kernel_params': {
            'gamma': 1.0,
            'n_components': 10,
            'num_threads': 1,
        },
        'ridge_params': {
            'regularizer': 1.0,
            'num_threads': 1,
            'fit_intercept': False
        }
    }

    For classification set 'objective' to 'logloss', and for regression use 'mse'.

    Parameters
    ----------

    params: dict

    Attributes
    ----------

    """

    def __init__(self,  params=None):

        self.ptr_ = np.array([], dtype=np.uint64)

        # params is a dictionary of dictionaries that includes all the model parameters 
        # params = {boosting_params, tree_params, rbf_params, ridge_params}
        # all *_params are dictionaries
        self.params_ = {}
        self.boosting_params_ = {}
        self.tree_params_ = {}
        self.kernel_params_ = {}
        self.ridge_params_ = {}
        
        # define boosting model defaults
        self.boosting_params_['num_round'] = 10
        self.boosting_params_['objective'] = 'mse'
        self.boosting_params_['learning_rate'] = 0.1
        self.boosting_params_['random_state'] = 0
        self.boosting_params_['colsample_bytree'] = 1.0
        self.boosting_params_['subsample'] = 1.0
        self.boosting_params_['verbose'] = False
        self.boosting_params_['enable_profile'] = False
        self.boosting_params_['lambda_l2'] = 0.0
        self.boosting_params_['early_stopping_rounds'] = 10
        self.boosting_params_['compress_trees'] = False
        self.boosting_params_['base_score'] = None
        self.params_['boosting_params'] = self.boosting_params_
        
        # define tree model defaults
        self.tree_params_['min_max_depth'] = 1
        self.tree_params_['max_max_depth'] = 6
        self.tree_params_['num_threads'] = 1
        self.tree_params_['use_histograms'] = True
        self.tree_params_['hist_nbins'] = 256
        self.tree_params_['use_gpu'] = False
        self.tree_params_['gpu_id'] = 0
        self.tree_params_['parallel_by_example'] = True
        self.tree_params_['select_probability'] = 1.0
        self.params_['tree_params'] = self.tree_params_

        # define ridge regression model defaults
        self.ridge_params_['regularizer'] = 1.0
        self.ridge_params_['num_threads'] = 1
        self.ridge_params_['fit_intercept'] = False
        self.ridge_params_['select_probability'] = 1.0 - self.tree_params_['select_probability']
        self.params_['ridge_params'] = self.ridge_params_

        # define kernel approximator (rbf sampler) defaults
        self.kernel_params_['gamma'] = 1.0
        self.kernel_params_['n_components'] = 10
        self.kernel_params_['num_threads'] = 1
        self.params_['kernel_params'] = self.kernel_params_
        
        if params is not None:
        
            params_keys = params.keys()
            
            # Check if the user provided keys are supported (first level)
            library_first_level_keys = list(self.params_.keys())
            unsupported_keys = list(np.setdiff1d(list(params_keys), library_first_level_keys))
            if len(unsupported_keys) > 0:
                raise KeyError("Unsupported keys in the user-defined parameters dictionary: ", unsupported_keys)
            
            # Check if the user provided keys are supported (second level)
            for params_type in library_first_level_keys:
                if params_type in params_keys:
                    user_keys = list(params[params_type].keys())
                    library_keys = list(self.params_[params_type].keys())
                    unsupported_keys = list(np.setdiff1d(user_keys, library_keys))
                    if len(unsupported_keys) > 0:
                        raise KeyError('Unsupported keys in the user-defined {} parameters: {}'.format(params_type, unsupported_keys))
            
            '''
            Boosting-model-specific parameters
            '''
            
            if 'boosting_params' in params_keys:
                boosting_keys = params['boosting_params'].keys()
            
                if 'num_round' in boosting_keys:
                    if params['boosting_params']['num_round'] <= 0:
                        raise ValueError("Parameter boosting_params:num_round should be an integer >= 1.")
                    self.params_['boosting_params']['num_round'] = int(params['boosting_params']['num_round'])

                if 'objective' in boosting_keys:
                    if params['boosting_params']['objective'] not in ['mse', 'logloss']:
                        raise ValueError("Parameter boosting_params:objective can be 'mse' or 'logloss'.")
                    self.params_['boosting_params']['objective'] = params['boosting_params']['objective']

                if 'learning_rate' in boosting_keys:
                    if params['boosting_params']['learning_rate'] < 0.0:
                        raise ValueError("Parameter boosting_params:learning_rate should be positive.")
                    self.params_['boosting_params']['learning_rate'] = float(params['boosting_params']['learning_rate'])

                if 'random_state' in boosting_keys:
                    if not isinstance(params['boosting_params']['random_state'], (int)):
                        raise ValueError("Parameter boosting_params:random_state : invalid type or value. Valid type is int.")
                    self.params_['boosting_params']['random_state'] = params['boosting_params']['random_state']

                if 'colsample_bytree' in boosting_keys:
                    if params['boosting_params']['colsample_bytree'] <= 0 or params['boosting_params']['colsample_bytree'] > 1:
                        raise ValueError("Parameter boosting_params:colsample_bytree : invalid value. Valid values are in (0,1].")
                    self.params_['boosting_params']['colsample_bytree'] = float(params['boosting_params']['colsample_bytree'])

                if 'subsample' in boosting_keys:
                    if params['boosting_params']['subsample'] <= 0 or params['boosting_params']['subsample'] > 1:
                        raise ValueError("Parameter boosting_params:subsample : invalid value. Valid values are in (0,1].")
                    self.params_['boosting_params']['subsample'] = float(params['boosting_params']['subsample'])

                if 'verbose' in boosting_keys:
                    if not isinstance(params['boosting_params']['verbose'], (bool)) :
                        raise TypeError("Parameter boosting_params:verbose : invalid type. Valid type is bool {True, False}.")
                    self.params_['boosting_params']['verbose'] = params['boosting_params']['verbose']

                if 'enable_profile' in boosting_keys:
                    if not isinstance(params['boosting_params']['enable_profile'], (bool)) :
                        raise TypeError("Parameter boosting_params:enable_profile : invalid type. Valid type is bool {True, False}.")
                    self.params_['boosting_params']['enable_profile'] = params['boosting_params']['enable_profile']

                if 'lambda_l2' in boosting_keys:
                    if params['boosting_params']['lambda_l2'] < 0:
                        raise ValueError("Parameter boosting_params:lambda_l2 should be a float >= 0.")
                    self.params_['boosting_params']['lambda_l2'] = float(params['boosting_params']['lambda_l2'])

                if 'early_stopping_rounds' in boosting_keys:
                    if params['boosting_params']['early_stopping_rounds'] < 1:
                        raise ValueError("Parameter boosting_params:early_stopping_rounds should be an integer >= 1.")
                    self.params_['boosting_params']['early_stopping_rounds'] = int(params['boosting_params']['early_stopping_rounds'])

                if 'compress_trees' in boosting_keys:
                    if not isinstance(params['boosting_params']['compress_trees'], (bool)) :
                        raise TypeError("Parameter boosting_params:compres_trees : invalid type. Valid type is bool {True, False}.")
                    self.params_['boosting_params']['compress_trees'] = params['boosting_params']['compress_trees']
                    
                if 'base_score' in boosting_keys:
                    self.params_['boosting_params']['base_score'] = params['boosting_params']['base_score']

            '''
            Tree-model-specific parameters
            '''
            
            if 'tree_params' in params_keys:
                tree_keys = params['tree_params'].keys()
                
                if 'min_max_depth' in tree_keys:
                    if params['tree_params']['min_max_depth'] <= 0:
                        raise ValueError("Parameter tree_params:min_max_depth should be an integer >= 1.")
                    self.params_['tree_params']['min_max_depth'] = int(params['tree_params']['min_max_depth'])

                if 'max_max_depth' in tree_keys:
                    if params['tree_params']['max_max_depth'] <= 0:
                        raise ValueError("Parameter tree_params:max_max_depth should be an integer >= 1.")
                    self.params_['tree_params']['max_max_depth'] = int(params['tree_params']['max_max_depth'])

                if self.params_['tree_params']['max_max_depth'] < self.params_['tree_params']['min_max_depth']:
                    raise ValueError("Parameter tree_params:max_max_depth should be >= tree_params_:min_max_depth.")

                if 'num_threads' in tree_keys:
                    if params['tree_params']['num_threads'] < 0:
                        raise ValueError("Parameter tree_params:n_threads should be an integer >= 1.")
                    self.params_['tree_params']['num_threads'] = int(params['tree_params']['num_threads'])

                if 'use_histograms' in tree_keys:
                    if not isinstance(params['tree_params']['use_histograms'], (bool)):
                        raise TypeError("Parameter tree_params:use_histograms : invalid type. Valid type is bool {True, False}.")
                    self.params_['tree_params']['use_histograms'] = params['tree_params']['use_histograms']

                if 'hist_nbins' in tree_keys:
                    if not isinstance(params['tree_params']['hist_nbins'], (int)):
                        raise TypeError("Parameter tree_params:hist_nbins : invalid type. Valid type is int > 1.")
                    if params['tree_params']['hist_nbins'] < 1 or params['tree_params']['hist_nbins'] > 256:
                        raise ValueError("Parameter tree_params:hist_nbins : invalid value. The number of bins must be between 1 and 256.")
                    self.params_['tree_params']['hist_nbins'] = int(params['tree_params']['hist_nbins'])

                if 'use_gpu' in tree_keys:
                    if not isinstance(params['tree_params']['use_gpu'], (bool)):
                        raise TypeError("Parameter tree_params:use_gpu : invalid type. Valid type is bool {True, False}.")
                    if params['tree_params']['use_gpu'] == True:
                        if 'use_histograms' in tree_keys:
                            if (params['tree_params']['use_histograms'] == False):
                                raise ValueError("GPU acceleration can only be enabled if tree_params:use_histograms parameter is True.")
                    self.params_['tree_params']['use_gpu'] = params['tree_params']['use_gpu']

                if 'gpu_id' in tree_keys:
                    if params['tree_params']['gpu_id'] < 0:
                        raise ValueError("Parameter tree_params:gpu_id should be an integer >= 0.")
                    self.params_['tree_params']['gpu_id'] = params['tree_params']['gpu_id']

                if 'parallel_by_example' in tree_keys:
                    if not isinstance(params['tree_params']['parallel_by_example'], (bool)):
                        raise TypeError("Parameter tree_params:parallel_by_example : invalid type. Valid type is bool {True, False}.")
                    self.params_['tree_params']['parallel_by_example'] = params['tree_params']['parallel_by_example']

                if 'select_probability' in tree_keys:
                    if params['tree_params']['select_probability'] < 0.0 or params['tree_params']['select_probability'] > 1.0:
                        raise ValueError("Parameter tree_params:select_probability : invalid value. Valid value is in [0,1].")
                    self.params_['tree_params']['select_probability'] = float(params['tree_params']['select_probability'])

            
            '''
            Ridge-regression-model-specific parameters
            '''
            
            if 'ridge_params' in params_keys:
                ridge_keys = params['ridge_params'].keys()

                if 'regularizer' in ridge_keys:
                    if params['ridge_params']['regularizer'] <= 0:
                        raise ValueError("Parameter ridge_params:regularizer : value should be positive.")
                    else:
                        self.params_['ridge_params']['regularizer'] = float(params['ridge_params']['regularizer'])

                if 'num_threads' in ridge_keys:
                    if params['ridge_params']['num_threads'] <= 0:
                        raise ValueError("Parameter ridge_params:num_threads : value should be positive.")
                    self.params_['ridge_params']['num_threads'] = int(params['ridge_params']['num_threads'])

                if 'fit_intercept' in ridge_keys:
                    if not isinstance(params['ridge_params']['fit_intercept'], (bool)):
                        raise ValueError("Parameter ridge_params:fit_intercept must be either False or True.")
                    self.params_['ridge_params']['fit_intercept'] = params['ridge_params']['fit_intercept']

                if 'select_probability' in ridge_keys:
                    if params['ridge_params']['select_probability'] < 0.0 or params['ridge_params']['select_probability'] > 1.0:
                        raise ValueError("Parameter ridge_params:select_probability : invalid value. Valid value is in [0,1].")
                    if (params['ridge_params']['select_probability'] + self.params_['tree_params']['select_probability'] != 1.0):
                        print("The sum of the tree and ridge selection probabilities should be 1.0. Updating probabilities proba(ridge) = 1 - proba(tree).")
                        self.params_['ridge_params']['select_probability'] = 1.0 - self.params_['tree_params']['select_probability']
                    else:    
                        self.params_['ridge_params']['select_probability'] = float(params['ridge_params']['select_probability'])
                else:
                    self.params_['ridge_params']['select_probability'] = float(1.0 - self.params_['tree_params']['select_probability'])

            '''
            RBF-Sampler-model-specific parameters
            '''

            if 'kernel_params' in params_keys:
                kernel_keys = params['kernel_params'].keys()

                if 'gamma' in kernel_keys:
                    self.params_['kernel_params']['gamma'] = float(params['kernel_params']['gamma'])

                if 'n_components' in kernel_keys:
                    if not isinstance(params['kernel_params']['n_components'], (int)):
                        raise TypeError("Parameter kernel_params:n_components : type invalid. It must be an integer >= 1.")
                    if params['kernel_params']['n_components'] < 1:
                        raise ValueError("Parameter kernel_params:n_components : value invalid. It must be an integer >= 1.")
                    self.params_['kernel_params']['n_components'] = int(params['kernel_params']['n_components'])

                if 'num_threads' in kernel_keys:
                    if params['kernel_params']['num_threads'] <= 0:
                        raise ValueError("Parameter kernel_params:num_threads : value should be positive.")
                    self.params_['kernel_params']['num_threads'] = int(params['kernel_params']['num_threads'])

        self.model_ = np.array([], dtype=np.float32)
        self.model_size_ = 0

    def __getstate__(self):
        # reset ptr
        self.ptr = np.array([], dtype=np.uint64)
        return self.__dict__

    def get_params(self):

        """
        Get the values of the model parameters.

        Returns
        -------
        params : dict
        """

        return self.params_

    def fit(self, X_train, y_train, sample_weight=None, X_val=None, y_val=None, sample_weight_val=None):

        """
        Fit the model according to the given train data.

        Parameters
        ----------
        X_train : dense matrix (ndarray)
            Train dataset

        y_train : array-like, shape = (n_samples,)
            The target vector corresponding to X_train.

        sample_weight : array-like, shape = (n_samples,)
            Training sample weights

        X_val : dense matrix (ndarray)
            Validation dataset

        y_val : array-like, shape = (n_samples,)
            The target vector corresponding to X_val.

        sample_weight_val : array-like, shape = (n_samples,)
            Validation sample weights

        Returns
        -------
        self : object
        """

        # Boosting Machine model random state
        #if (self.params_['boosting_params']['random_state'] == None):
        #    # Not sure if this should be the random state
        #    random_state = np.random.get_state()[1][0]
        #else:
        #random_state = self.params_['boosting_params']['random_state']

        # the user has not set the base score, thus we will set it so that it speeds up the learning
        if self.params_['boosting_params']['base_score'] is None:
        
            # this is a regression problem        
            if self.params_['boosting_params']['objective'] == 'mse':
                if sample_weight is None:
                    self.params_['boosting_params']['base_score'] = np.average(y_train)
                else:
                    self.params_['boosting_params']['base_score'] = np.average(y_train, weights=sample_weight)
        
            # this is a classification problem
            else:  
                if sample_weight is None:
                    sum_positives = np.sum(y_train > 0)
                    sum_negatives = y_train.shape[0] - sum_positives
                    self.params_['boosting_params']['base_score'] = np.log(sum_positives/sum_negatives) if sum_negatives > 0 and sum_positives > 0 else 0.0
                else:
                    sum_positives = np.sum(sample_weight[y_train > 0])
                    sum_negatives = np.sum(sample_weight) - sum_positives
                    self.params_['boosting_params']['base_score'] = np.log(sum_positives/sum_negatives) if sum_negatives > 0 and sum_positives > 0 else 0.0
                    
        if type(X_train).__name__ != "ndarray":
            raise TypeError("Tree-based models in Snap ML only support numpy.ndarray")

        if X_val is not None and type(X_val).__name__ != "ndarray":
            raise TypeError("Tree-based models in Snap ML only support numpy.ndarray")

        # helper function to prep data
        def prep_data(X, y, name):
            num_ft = 0
            num_nz = 0
            indptr = np.array([])
            indices = np.array([])
            data = np.array([])
            labs = np.array([])

            num_ex = len(y)

            # Check train dataset format
            if (X.shape[0] != num_ex):
                raise ValueError("X_%s.shape[0] must be equal to number of examples." % (name))

            # Keep the sparsity checks for future support for sparse datasets
            if (type(X).__name__ == "csr_matrix"):
                num_ft = X.shape[1]
            elif (type(X).__name__ == "csc_matrix"):
                X_train = X.tocsr()
                num_ft = X.shape[1]
            elif (type(X).__name__ == "ndarray"):
                if (X.shape[0] == X.shape[1]):
                    print("[Warning] Each row in the input %s dataset is assumed to be an example." % (name))
                    num_ft = X.shape[0]
                elif (X.shape[0] == num_ex):
                    num_ft = X.shape[1]
                else:
                    X = X.transpose().copy()
                    num_ft = X.shape[1]

            if (type(X).__name__ == "csr_matrix" or type(X).__name__ == "csc_matrix"):
                num_nz    = X.nnz
                indptr    = X.indptr.astype(np.uint64)
                data      = X.data.astype(np.float32)
                indices   = X.indices.astype(np.uint32)
            elif (type(X).__name__ == "ndarray"):
                num_nz    = num_ex * num_ft
                data      = X.astype(np.float32)
                data      = np.ascontiguousarray(data)
                indptr    = np.array([])
                indices   = np.array([])
            else:
                raise TypeError("X_%s should be in csr_matrix, csc_matrix or ndarray format." % (name))

            labs = y.astype(np.float32)

            return num_ex, num_ft, num_nz, indptr, indices, data, labs

        # prepare training data
        train_num_ex, train_num_ft, train_num_nz, train_indptr, train_indices, train_data, train_labs = prep_data(X_train, y_train, "train")

        # prepare validation data
        if X_val is not None and y_val is not None:
            val_num_ex, val_num_ft, val_num_nz, val_indptr, val_indices, val_data, val_labs = prep_data(X_val, y_val, "val")
        else:
            val_num_ex, val_num_ft, val_num_nz, val_indptr, val_indices, val_data, val_labs = 0, 0, 0, np.array([]), np.array([]), np.array([]), np.array([])

        if not sample_weight is None:
            if type(sample_weight).__name__ != "ndarray":
                raise TypeError("Parameter sample_weight: invalid type. Supported type: ndarray.")
            sample_weight = sample_weight.astype(np.float32)
        else:
            sample_weight = np.array([], dtype=np.float32)

        if not sample_weight_val is None:
            if type(sample_weight_val).__name__ != "ndarray":
                raise TypeError("Parameter sample_weight_val: invalid type. Supported type: ndarray.")
            if X_val is None:
                raise ValueError("Parameter sample_weight_val not supported when X_val and y_val are not defined.")
            sample_weight_val = sample_weight_val.astype(np.float32)
        else:
            sample_weight_val = np.array([], dtype=np.float32)


        out_ = libsnapml.booster_fit(self.params_['boosting_params'],
                                     self.params_['tree_params'],
                                     self.params_['ridge_params'],
                                     self.params_['kernel_params'],
                                     train_num_ex, train_num_ft, train_num_nz,
                                     train_indptr, train_indices, train_data, train_labs,
                                     val_num_ex, val_num_ft, val_num_nz,
                                     val_indptr, val_indices, val_data, val_labs,
                                     sample_weight, sample_weight_val)

        self.model_ = out_[0]
        self.model_size_ = out_[1]['model_size']
        self.ensemble_size_ = out_[1]['best_num_rounds']

        return None

    def predict(self, X, num_threads=1):

        """
        Raw predictions

        If the training objective is 'mse' then it returns the predicted estimates.
        If the training objective is 'logloss' then it returns the predicted estimates
        before the logistic transformation (raw logits).

        Parameters
        ----------
        X : dense matrix (ndarray)
            Dataset used for predicting class estimates.

        num_threads : int
            Number of threads to use for prediction.

        Returns
        -------
        pred: array-like, shape = (n_samples,)
            Returns the predicted estimates.
        """

        if type(X).__name__ != "ndarray":
            raise TypeError("Tree-based models in Snap ML only support numpy.ndarray")

        num_ex = 0
        num_ft = 0
        indptr = np.array([])
        indices =  np.array([])
        data = np.array([])

        num_ex = X.shape[0]
        num_ft = X.shape[1]

        # Keep the sparsity checks for future support for sparse datasets
        if (type(X).__name__ == "csc_matrix"):
            X = X.tocsr()
            indptr    = X.indptr.astype(np.uint64)
            data      = X.data.astype(np.float32)
            indices   = X.indices.astype(np.uint32)
        elif (type(X).__name__ == "csr_matrix"):
            indptr    = X.indptr.astype(np.uint64)
            data      = X.data.astype(np.float32)
            indices   = X.indices.astype(np.uint32)
        elif (type(X).__name__ == "ndarray"):
            data      = X.astype(np.float32)
            data      = np.ascontiguousarray(data)    # enforce row-major format
            indptr    = np.array([])
            indices   = np.array([])
        else:
            raise ValueError("X should be in csc_matrix, csr_matrix or ndarray format.")

        pred = []

        # flag that indicates if output probabilities are to be computed
        # 0 : get raw predictions
        # 1 : get output probabilities (only for predict proba)
        get_proba = 0

        # Generate predictions
        pred, ptr = libsnapml.booster_predict(num_ex,
                                         num_ft,
                                         indptr,
                                         indices,
                                         data,
                                         self.params_['boosting_params']['num_round'],
                                         self.params_['boosting_params']['learning_rate'],
                                         self.params_['boosting_params']['colsample_bytree'],
                                         self.params_['kernel_params']['n_components'],
                                         self.params_['boosting_params']['random_state'],
                                         self.params_['boosting_params']['base_score'],
                                         self.model_,
                                         self.model_size_,
                                         get_proba,
                                         self.params_['ridge_params']['fit_intercept'],
                                         self.params_['kernel_params']['gamma'],
                                         num_threads,
                                         self.ptr_)

        if ptr is not None:
            self.ptr_ = ptr

        pred = np.array(pred)

        if not np.all(np.isfinite(pred)):
            warnings.warn("Boosting diverged; Try using a smaller learning rate.")

        # handle case of divergence
        pred = np.nan_to_num(pred)

        return pred


    def predict_proba(self, X, num_threads=1):

        """
        Output probabilities

        Use only if the training objective is 'logloss' (i.e., for binary classification problems).
        It returns the probabilities of each sample belonging to each class. 
        The probabilities are calculated using the logistic transformation. 

        Parameters
        ----------
        X : dense matrix (ndarray)
            Dataset used for predicting class estimates.

        num_threads : int
            Number of threads to use for prediction.

        Returns
        -------
        proba: array-like, shape = (n_samples, 2)
            Returns the predicted probabilities of each sample belonging to each class.
        """

        if type(X).__name__ != "ndarray":
            raise TypeError("Tree-based models in Snap ML only support numpy.ndarray")

        num_ex = 0
        num_ft = 0
        indptr = np.array([])
        indices =  np.array([])
        data = np.array([])

        num_ex = X.shape[0]
        num_ft = X.shape[1]

        # Keep the sparsity checks for future support for sparse datasets
        if (type(X).__name__ == "csc_matrix"):
            X = X.tocsr()
            indptr    = X.indptr.astype(np.uint64)
            data      = X.data.astype(np.float32)
            indices   = X.indices.astype(np.uint32)
        elif (type(X).__name__ == "csr_matrix"):
            indptr    = X.indptr.astype(np.uint64)
            data      = X.data.astype(np.float32)
            indices   = X.indices.astype(np.uint32)
        elif (type(X).__name__ == "ndarray"):
            data      = X.astype(np.float32)
            data      = np.ascontiguousarray(data)    # enforce row-major format
            indptr    = np.array([])
            indices   = np.array([])
        else:
            raise ValueError("X should be in csc_matrix, csr_matrix or ndarray format.")

        pred = []

        # flag that indicates if output probabilities are to be computed
        # 0 : get raw predictions
        # 1 : get output probabilities (only for predict proba)
        get_proba = 1

        # Generate predictions
        pred, ptr = libsnapml.booster_predict(num_ex,
                                         num_ft,
                                         indptr,
                                         indices,
                                         data,
                                         self.params_['boosting_params']['num_round'],
                                         self.params_['boosting_params']['learning_rate'],
                                         self.params_['boosting_params']['colsample_bytree'],
                                         self.params_['kernel_params']['n_components'],
                                         self.params_['boosting_params']['random_state'],
                                         self.params_['boosting_params']['base_score'],
                                         self.model_,
                                         self.model_size_,
                                         get_proba,
                                         self.params_['ridge_params']['fit_intercept'],
                                         self.params_['kernel_params']['gamma'],
                                         num_threads,
                                         self.ptr_)

        if ptr is not None:
            self.ptr_ = ptr

        pred = np.array(pred)

        if not np.all(np.isfinite(pred)):
            warnings.warn("Boosting diverged; Try using a smaller learning rate.")

         # handle case of divergence
        pred = np.nan_to_num(pred)

        return pred


    def compress_trees(self, X):

        """
        Compress decision trees for fast inference

        Parameters
        ----------
        X : dense matrix (ndarray)
            Dataset used for compressing trees

        """

        num_ex = X.shape[0]
        num_ft = X.shape[1]

        # Validate data type
        if (type(X).__name__ == "ndarray"):
            data = X.astype(np.float32)
            data = np.ascontiguousarray(data)    # enforce row-major format
        else:
            raise ValueError("X should be in ndarray format.")

        # Compress trees
        out = libsnapml.booster_compress(num_ex,
                                         num_ft,
                                         data,
                                         self.params_['boosting_params']['num_round'],
                                         self.model_,
                                         self.model_size_)

        self.model_ = out[0]
        self.model_size_ = out[1]['model_size']
