# *****************************************************************
#
# Licensed Materials - Property of IBM
#
# (C) Copyright IBM Corp. 2019, 2020. All Rights Reserved.
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
#
# *****************************************************************

from __future__ import print_function

import time

# Snap ML APIs
from snapml import MPI
from snapml.io import load_from_snap_format
from snapml import SupportVectorMachine
from snapml.metrics import mean_squared_error as snapMSE, accuracy_score as snapACC

# sklearn APIs
from sklearn.datasets import load_svmlight_file
from sklearn.svm import LinearSVC
from sklearn.metrics import accuracy_score, mean_squared_error

(comm_size, rank) = MPI.Comm_get_info()

import argparse

defaultPath = "."
CLI=argparse.ArgumentParser()
CLI.add_argument(
   "--data_path",
   type=str,
   default=defaultPath
)
CLI.add_argument(
   "--device_ids",
   nargs="*",
   type=int,
   default=[]
)
CLI.add_argument(
    "--use_gpu",
    action = 'store_true',
    default=False
)
CLI.add_argument(
    "--skip_sklearn",
    action = 'store_true',
    default=False
)
CLI.add_argument(
    "--use_balanced_class_weights",
    action = 'store_true',
    default=False
)

args = CLI.parse_args()

if args.use_balanced_class_weights:
  class_weight = "balanced"
else:
  class_weight = None

data = load_from_snap_format(args.data_path + "/data/news20.train.snap")

# Create a Support Vector Machine (setting GPU appropriately)
svm = SupportVectorMachine(device_ids=args.device_ids, use_gpu=args.use_gpu,class_weight=class_weight)

# Training
t0 = time.time()
svm.fit(data)

if (rank == 0):
    print("[snapml] MPI process %d : Training time (s):  %.2f" % (rank, time.time()-t0))

# Inference
pred_class = svm.predict(data)

# Evaluate log-loss on test set
acc = snapACC(data, pred_class)
mse = snapMSE(data, pred_class)
if (rank == 0):
    print("[snapml] Accuracy:                 %.4f" % (acc))
    print("[snapml] Mean squared error:       %.4f" % (mse))

if args.skip_sklearn:
    if (rank == 0):
        print('Execution using scikit-learn is skipped.')
    exit()

X,y = load_svmlight_file(args.data_path + "/data/news20.binary")

# Create a LogisticRegression (dual is faster for this application)
clf = LinearSVC(dual=True, fit_intercept=False, loss="hinge")

# Training time
t0 = time.time()
clf.fit(X, y)

if (rank == 0):
    print("[sklearn] Training time (s):  %.2f" % (time.time()-t0))

# Inference
pred_class = clf.predict(X)

if (rank == 0):
    print("[sklearn] Accuracy:              %.4f" % (accuracy_score(y, pred_class)))
    print("[sklearn] Mean squared error:    %.4f" % (mean_squared_error(y, pred_class)))
